/* This file was generated automatically by genconf.c. */
#ifdef oper_
oper_(zbseq_l2_op_defs)
#endif
#ifdef psfile_
psfile_("gs_btokn.ps",11)
#endif
#ifdef oper_
oper_(zcolor1_op_defs)
oper_(zht1_op_defs)
oper_(zupath_l2_op_defs)
oper_(zupath_op_defs)
oper_(zvmem2_op_defs)
oper_(zdps1_l2_op_defs)
#endif
#ifdef psfile_
psfile_("gs_dps1.ps",10)
#endif
#ifdef oper_
oper_(ireclaim_l2_op_defs)
oper_(zcharx_op_defs)
#endif
#ifdef psfile_
psfile_("gs_dps2.ps",10)
#endif
#ifdef oper_
oper_(zchar1_op_defs)
oper_(zfont1_op_defs)
oper_(zmisc1_op_defs)
#endif
#ifdef psfile_
psfile_("gs_agl.ps",9)
psfile_("gs_type1.ps",11)
#endif
#ifdef oper_
oper_(zusparam_op_defs)
oper_(zmisc2_op_defs)
#endif
#ifdef psfile_
psfile_("gs_lev2.ps",10)
psfile_("gs_res.ps",9)
#endif
#ifdef oper_
oper_(zchar2_op_defs)
oper_(zfont2_op_defs)
oper_(zchar42_op_defs)
oper_(zfont42_op_defs)
#endif
#ifdef psfile_
psfile_("gs_typ42.ps",11)
#endif
#ifdef oper_
oper_(zfrsd_op_defs)
oper_(zfcid0_op_defs)
oper_(zfcid1_op_defs)
#endif
#ifdef psfile_
psfile_("gs_cidfn.ps",11)
psfile_("gs_cidcm.ps",11)
psfile_("gs_fntem.ps",11)
psfile_("gs_cidtt.ps",11)
psfile_("gs_cidfm.ps",11)
#endif
#ifdef oper_
oper_(zcrd_l2_op_defs)
oper_(zfcmap_op_defs)
#endif
#ifdef psfile_
psfile_("gs_cmap.ps",10)
#endif
#ifdef oper_
oper_(zcfont_op_defs)
oper_(zfont0_op_defs)
oper_(zfdcte_op_defs)
oper_(zfdctd_op_defs)
oper_(zfdecode_op_defs)
oper_(zfilter2_op_defs)
oper_(ziodev2_l2_op_defs)
#endif
#ifdef io_device_
io_device_(gs_iodev_null)
io_device_(gs_iodev_calendar)
#endif
#ifdef oper_
oper_(zdevice2_l2_op_defs)
oper_(zmedia2_l2_op_defs)
#endif
#ifdef psfile_
psfile_("gs_setpd.ps",11)
#endif
#ifdef image_type_
image_type_(3,gs_image_type_3)
image_type_(4,gs_image_type_4)
#endif
#ifdef oper_
oper_(zpcolor_l2_op_defs)
oper_(zarith_op_defs)
oper_(zarray_op_defs)
oper_(zcontrol1_op_defs)
oper_(zcontrol2_op_defs)
oper_(zcontrol3_op_defs)
oper_(zdict1_op_defs)
oper_(zdict2_op_defs)
oper_(zfile_op_defs)
oper_(zfile1_op_defs)
oper_(zfileio1_op_defs)
oper_(zfileio2_op_defs)
oper_(zfilter_op_defs)
oper_(zfproc_op_defs)
oper_(zgeneric_op_defs)
oper_(ziodev_op_defs)
oper_(zmath_op_defs)
oper_(zalg_op_defs)
oper_(zmisc_a_op_defs)
oper_(zmisc_b_op_defs)
oper_(zpacked_op_defs)
oper_(zrelbit_op_defs)
oper_(zstack_op_defs)
oper_(zstring_op_defs)
oper_(zsysvm_op_defs)
oper_(ztoken_op_defs)
oper_(ztype_op_defs)
oper_(zvmem_op_defs)
oper_(zbfont_op_defs)
oper_(zchar_a_op_defs)
oper_(zchar_b_op_defs)
oper_(zcolor_op_defs)
oper_(zcolor_ext_op_defs)
oper_(zdevice_op_defs)
oper_(zdevice_ext_op_defs)
oper_(zfont_op_defs)
oper_(zfontenum_op_defs)
oper_(zgstate1_op_defs)
oper_(zgstate2_op_defs)
oper_(zgstate3_op_defs)
oper_(zht_op_defs)
oper_(zimage_op_defs)
oper_(zmatrix_op_defs)
oper_(zmatrix2_op_defs)
oper_(zpaint_op_defs)
oper_(zpath_op_defs)
oper_(pantone_op_defs)
#endif
#ifdef io_device_
io_device_(gs_iodev_stdin)
io_device_(gs_iodev_stdout)
io_device_(gs_iodev_stderr)
io_device_(gs_iodev_lineedit)
io_device_(gs_iodev_statementedit)
#endif
#ifdef oper_
oper_(zdscpars_op_defs)
oper_(zfapi_op_defs)
#endif
#ifdef psfile_
psfile_("gs_fapi.ps",10)
#endif
#ifdef oper_
oper_(zfbcp_op_defs)
oper_(zpath1_op_defs)
#endif
#ifdef emulator_
emulator_("PostScript",10)
emulator_("PostScriptLevel1",16)
#endif
#ifdef oper_
oper_(zht2_l2_op_defs)
oper_(zcssepr_l2_op_defs)
oper_(zfsample_op_defs)
#endif
#ifdef function_type_
function_type_(4,gs_build_function_4)
#endif
#ifdef oper_
oper_(zfunc_op_defs)
#endif
#ifdef function_type_
function_type_(0,gs_build_function_0)
#endif
#ifdef oper_
oper_(zchar32_op_defs)
oper_(zfont32_op_defs)
#endif
#ifdef psfile_
psfile_("gs_typ32.ps",11)
#endif
#ifdef oper_
oper_(zfimscale_op_defs)
oper_(zform_op_defs)
#endif
#ifdef emulator_
emulator_("PostScriptLevel2",16)
#endif
#ifdef psfile_
psfile_("gs_frsd.ps",10)
#endif
#ifdef oper_
oper_(zimage3_op_defs)
oper_(zmisc3_op_defs)
oper_(zcolor3_l3_op_defs)
oper_(zshade_op_defs)
#endif
#ifdef function_type_
function_type_(2,gs_build_function_2)
function_type_(3,gs_build_function_3)
#endif
#ifdef psfile_
psfile_("gs_ll3.ps",9)
#endif
#ifdef oper_
oper_(zfzlib_op_defs)
oper_(zicc_op_defs)
#endif
#ifdef psfile_
psfile_("gs_icc.ps",9)
#endif
#ifdef emulator_
emulator_("PostScriptLevel3",16)
#endif
#ifdef psfile_
psfile_("gs_mex_e.ps",11)
psfile_("gs_mro_e.ps",11)
psfile_("gs_pdf_e.ps",11)
psfile_("gs_wan_e.ps",11)
#endif
#ifdef oper_
oper_(ztrans1_op_defs)
oper_(ztrans2_op_defs)
oper_(ztrans3_op_defs)
#endif
#ifdef image_type_
image_type_(3x,gs_image_type_3x)
#endif
#ifdef init_
init_(gs_gxclfile_init)
init_(gs_gxclmem_init)
#endif
#ifdef psfile_
psfile_("pdf_main.ps",11)
#endif
#ifdef oper_
oper_(zfmd5_op_defs)
oper_(zfsha2_op_defs)
oper_(zfarc4_op_defs)
oper_(zfaes_op_defs)
oper_(zpdfops_op_defs)
oper_(zpdf_r6_op_defs)
#endif
#ifdef psfile_
psfile_("gs_cff.ps",9)
psfile_("gs_mgl_e.ps",11)
psfile_("gs_ttf.ps",9)
#endif
#ifdef emulator_
emulator_("PDF",3)
#endif
#ifdef psfile_
psfile_("gs_epsf.ps",10)
#endif
#ifdef oper_
oper_(zfjpx_op_defs)
oper_(zfjbig2_op_defs)
#endif
#ifdef psfile_
psfile_("gs_pdfwr.ps",11)
#endif
#ifdef image_class_
image_class_(gs_image_class_2_fracs)
image_class_(gs_image_class_4_color)
#endif

