/* Copyright (C) 2001-2023 Artifex Software, Inc.
   All Rights Reserved.

   This software is provided AS-IS with no warranty, either express or
   implied.

   This software is distributed under license and may not be copied,
   modified or distributed except as expressly authorized under the terms
   of the license contained in the file LICENSE in this distribution.

   Refer to licensing information at http://www.artifex.com or contact
   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
   CA 94129, USA, for further information.
*/
/*
 * This file contains substantial parts of toolbin/encs2c.ps,
 * which generated the remainder of the file mechanically from
 *   gs_std_e.ps  gs_il1_e.ps  gs_sym_e.ps  gs_dbt_e.ps
 *   gs_wan_e.ps  gs_mro_e.ps  gs_mex_e.ps  gs_mgl_e.ps
 *   gs_lgo_e.ps  gs_lgx_e.ps  gs_css_e.ps
 *
 * This source file is maintained manually under source code control,
 * however its content should be regenerated by using encs2c.ps
 * if changes are required.
 * You should not manually alter this file! If you regenerate it using
 * encs2c.ps you must regenerate all 4 files; base/gscedata.[c|h]
 * and devices/vector/gdevpdtv.[c|h]
 */

#ifndef gdevpdtv_INCLUDED
#define gdevpdtv_INCLUDED

#define GS_C_PDF_MAX_GOOD_GLYPH 22086
#define GS_C_PDF_GOOD_GLYPH_MASK 1
#define GS_C_PDF_GOOD_NON_SYMBOL_MASK 2

extern const unsigned char gs_c_pdf_glyph_type[];

#endif /* gdevpdtv_INCLUDED */
