/* This file was generated automatically by genconf.c. */
#ifdef compositor_
compositor_(gs_composite_overprint_type)
compositor_(gs_composite_pdf14trans_type)
#endif
#ifdef device_
device_(gs_display_device)
#endif
#ifdef device2_
device2_(gs_PCLm_device)
device2_(gs_PCLm8_device)
device2_(gs_bbox_device)
device2_(gs_bit_device)
device2_(gs_bitcmyk_device)
device2_(gs_bitrgb_device)
device2_(gs_bitrgbtags_device)
device2_(gs_bmp16_device)
device2_(gs_bmp16m_device)
device2_(gs_bmp256_device)
device2_(gs_bmp32b_device)
device2_(gs_bmpgray_device)
device2_(gs_bmpmono_device)
device2_(gs_bmpsep1_device)
device2_(gs_bmpsep8_device)
#endif
#ifdef device_
device_(gs_ccr_device)
device_(gs_cfax_device)
device_(gs_cif_device)
device_(gs_devicen_device)
device_(gs_dfaxhigh_device)
device_(gs_dfaxlow_device)
#endif
#ifdef device2_
device2_(gs_docxwrite_device)
device2_(gs_eps2write_device)
device2_(gs_pdfwrite_device)
device2_(gs_faxg3_device)
device2_(gs_faxg32d_device)
device2_(gs_faxg4_device)
device2_(gs_fpng_device)
#endif
#ifdef device_
device_(gs_inferno_device)
#endif
#ifdef device2_
device2_(gs_ink_cov_device)
device2_(gs_inkcov_device)
device2_(gs_jpeg_device)
device2_(gs_jpegcmyk_device)
device2_(gs_jpeggray_device)
#endif
#ifdef device_
device_(gs_mgr4_device)
device_(gs_mgr8_device)
device_(gs_mgrgray2_device)
device_(gs_mgrgray4_device)
device_(gs_mgrgray8_device)
device_(gs_mgrmono_device)
device_(gs_miff24_device)
#endif
#ifdef device2_
device2_(gs_pam_device)
device2_(gs_pamcmyk32_device)
device2_(gs_pamcmyk4_device)
device2_(gs_pbm_device)
device2_(gs_pbmraw_device)
device2_(gs_pcx16_device)
device2_(gs_pcx24b_device)
device2_(gs_pcx256_device)
device2_(gs_pcxcmyk_device)
device2_(gs_pcxgray_device)
device2_(gs_pcxmono_device)
device2_(gs_pdfimage24_device)
device2_(gs_pdfimage32_device)
device2_(gs_pdfimage8_device)
device2_(gs_pdfwrite_device)
device2_(gs_pgm_device)
device2_(gs_pgmraw_device)
device2_(gs_pgnm_device)
device2_(gs_pgnmraw_device)
device2_(gs_pkm_device)
device2_(gs_pkmraw_device)
device2_(gs_pksm_device)
device2_(gs_pksmraw_device)
device2_(gs_plan_device)
device2_(gs_plan9bm_device)
device2_(gs_planc_device)
device2_(gs_plang_device)
device2_(gs_plank_device)
device2_(gs_planm_device)
device2_(gs_plib_device)
device2_(gs_plibc_device)
device2_(gs_plibg_device)
device2_(gs_plibk_device)
device2_(gs_plibm_device)
device2_(gs_pnm_device)
device2_(gs_pnmraw_device)
device2_(gs_ppm_device)
device2_(gs_ppmraw_device)
device2_(gs_pppm_device)
device2_(gs_ps2write_device)
device2_(gs_pdfwrite_device)
#endif
#ifdef device_
device_(gs_psdcmyk_device)
device_(gs_psdcmyk16_device)
device_(gs_psdcmykog_device)
device_(gs_psdcmyktags_device)
device_(gs_psdcmyktags16_device)
device_(gs_psdrgb_device)
device_(gs_psdrgb16_device)
device_(gs_psdrgbtags_device)
device_(gs_spotcmyk_device)
#endif
#ifdef device2_
device2_(gs_tiff12nc_device)
device2_(gs_tiff24nc_device)
device2_(gs_tiff32nc_device)
device2_(gs_tiff48nc_device)
device2_(gs_tiff64nc_device)
device2_(gs_tiffcrle_device)
device2_(gs_tiffg3_device)
device2_(gs_tiffg32d_device)
device2_(gs_tiffg4_device)
device2_(gs_tiffgray_device)
device2_(gs_tifflzw_device)
device2_(gs_tiffpack_device)
device2_(gs_tiffscaled_device)
device2_(gs_tiffscaled24_device)
device2_(gs_tiffscaled32_device)
device2_(gs_tiffscaled4_device)
device2_(gs_tiffscaled8_device)
device2_(gs_tiffsep_device)
device2_(gs_tiffsep1_device)
device2_(gs_txtwrite_device)
device2_(gs_urfcmyk_device)
device2_(gs_urfgray_device)
device2_(gs_urfrgb_device)
#endif
#ifdef device_
device_(gs_xcf_device)
#endif
#ifdef device2_
device2_(gs_xpswrite_device)
#endif
#ifdef device_
device_(gs_alc1900_device)
device_(gs_alc2000_device)
device_(gs_alc4000_device)
device_(gs_alc4100_device)
device_(gs_alc8500_device)
device_(gs_alc8600_device)
device_(gs_alc9100_device)
device_(gs_ap3250_device)
device_(gs_appledmp_device)
#endif
#ifdef device2_
device2_(gs_atx23_device)
device2_(gs_atx24_device)
device2_(gs_atx38_device)
#endif
#ifdef device_
device_(gs_bj10e_device)
device_(gs_bj10v_device)
device_(gs_bj10vh_device)
device_(gs_bj200_device)
device_(gs_bjc600_device)
device_(gs_bjc800_device)
device_(gs_bjc880j_device)
device_(gs_bjccmyk_device)
device_(gs_bjccolor_device)
device_(gs_bjcgray_device)
device_(gs_bjcmono_device)
device_(gs_cdeskjet_device)
#endif
#ifdef device2_
device2_(gs_cdj1600_device)
#endif
#ifdef device_
device_(gs_cdj500_device)
device_(gs_cdj550_device)
#endif
#ifdef device2_
device2_(gs_cdj670_device)
device2_(gs_cdj850_device)
device2_(gs_cdj880_device)
device2_(gs_cdj890_device)
device2_(gs_cdj970_device)
#endif
#ifdef device_
device_(gs_cdjcolor_device)
device_(gs_cdjmono_device)
#endif
#ifdef device2_
device2_(gs_cdnj500_device)
device2_(gs_chp2200_device)
#endif
#ifdef device_
device_(gs_cljet5_device)
device_(gs_cljet5c_device)
device_(gs_cljet5pr_device)
device_(gs_coslw2p_device)
device_(gs_coslwxl_device)
device_(gs_declj250_device)
#endif
#ifdef device2_
device2_(gs_deskjet_device)
#endif
#ifdef device_
device_(gs_dj505j_device)
#endif
#ifdef device2_
device2_(gs_djet500_device)
#endif
#ifdef device_
device_(gs_djet500c_device)
device_(gs_dl2100_device)
device_(gs_dnj650c_device)
device_(gs_epl2050_device)
device_(gs_epl2050p_device)
device_(gs_epl2120_device)
device_(gs_epl2500_device)
device_(gs_epl2750_device)
device_(gs_epl5800_device)
device_(gs_epl5900_device)
device_(gs_epl6100_device)
device_(gs_epl6200_device)
device_(gs_eplcolor_device)
device_(gs_eplmono_device)
device_(gs_eps9high_device)
device_(gs_eps9mid_device)
device_(gs_epson_device)
device_(gs_epsonc_device)
device_(gs_escp_device)
device_(gs_escpage_device)
device_(gs_fmlbp_device)
device_(gs_fmpr_device)
#endif
#ifdef device2_
device2_(gs_fs600_device)
#endif
#ifdef device_
device_(gs_gdi_device)
device_(gs_hl1240_device)
device_(gs_hl1250_device)
device_(gs_hl7x0_device)
device_(gs_hpdj1120c_device)
device_(gs_hpdj310_device)
device_(gs_hpdj320_device)
device_(gs_hpdj340_device)
device_(gs_hpdj400_device)
device_(gs_hpdj500_device)
device_(gs_hpdj500c_device)
device_(gs_hpdj510_device)
device_(gs_hpdj520_device)
device_(gs_hpdj540_device)
device_(gs_hpdj550c_device)
device_(gs_hpdj560c_device)
device_(gs_hpdj600_device)
device_(gs_hpdj660c_device)
device_(gs_hpdj670c_device)
device_(gs_hpdj680c_device)
device_(gs_hpdj690c_device)
device_(gs_hpdj850c_device)
device_(gs_hpdj855c_device)
device_(gs_hpdj870c_device)
device_(gs_hpdj890c_device)
device_(gs_hpdjplus_device)
device_(gs_hpdjportable_device)
device_(gs_ibmpro_device)
device_(gs_imagen_device)
#endif
#ifdef device2_
device2_(gs_itk24i_device)
device2_(gs_itk38_device)
#endif
#ifdef device_
device_(gs_iwhi_device)
device_(gs_iwlo_device)
device_(gs_iwlq_device)
device_(gs_jetp3852_device)
device_(gs_jj100_device)
device_(gs_la50_device)
device_(gs_la70_device)
device_(gs_la75_device)
device_(gs_la75plus_device)
#endif
#ifdef device2_
device2_(gs_laserjet_device)
#endif
#ifdef device_
device_(gs_lbp310_device)
device_(gs_lbp320_device)
device_(gs_lbp8_device)
device_(gs_lex2050_device)
device_(gs_lex3200_device)
device_(gs_lex5700_device)
device_(gs_lex7000_device)
device_(gs_lips2p_device)
device_(gs_lips3_device)
device_(gs_lips4_device)
device_(gs_lips4v_device)
device_(gs_lj250_device)
device_(gs_lj3100sw_device)
device_(gs_lj4dith_device)
device_(gs_lj4dithp_device)
device_(gs_lj5gray_device)
device_(gs_lj5mono_device)
#endif
#ifdef device2_
device2_(gs_ljet2p_device)
device2_(gs_ljet3_device)
device2_(gs_ljet3d_device)
device2_(gs_ljet4_device)
device2_(gs_ljet4d_device)
#endif
#ifdef device_
device_(gs_ljet4pjl_device)
#endif
#ifdef device2_
device2_(gs_ljetplus_device)
#endif
#ifdef device_
device_(gs_ln03_device)
device_(gs_lp1800_device)
device_(gs_lp1900_device)
device_(gs_lp2000_device)
device_(gs_lp2200_device)
device_(gs_lp2400_device)
device_(gs_lp2500_device)
#endif
#ifdef device2_
device2_(gs_lp2563_device)
#endif
#ifdef device_
device_(gs_lp3000c_device)
device_(gs_lp7500_device)
device_(gs_lp7700_device)
device_(gs_lp7900_device)
device_(gs_lp8000_device)
device_(gs_lp8000c_device)
device_(gs_lp8100_device)
device_(gs_lp8200c_device)
device_(gs_lp8300c_device)
device_(gs_lp8300f_device)
device_(gs_lp8400f_device)
device_(gs_lp8500c_device)
device_(gs_lp8600_device)
device_(gs_lp8600f_device)
device_(gs_lp8700_device)
device_(gs_lp8800c_device)
device_(gs_lp8900_device)
device_(gs_lp9000b_device)
device_(gs_lp9000c_device)
device_(gs_lp9100_device)
device_(gs_lp9200b_device)
device_(gs_lp9200c_device)
device_(gs_lp9300_device)
device_(gs_lp9400_device)
device_(gs_lp9500c_device)
device_(gs_lp9600_device)
device_(gs_lp9600s_device)
device_(gs_lp9800c_device)
device_(gs_lps4500_device)
device_(gs_lps6500_device)
device_(gs_lq850_device)
device_(gs_lxm3200_device)
device_(gs_lxm5700m_device)
device_(gs_m8510_device)
device_(gs_md1xMono_device)
device_(gs_md2k_device)
device_(gs_md50Eco_device)
device_(gs_md50Mono_device)
device_(gs_md5k_device)
device_(gs_mj500c_device)
device_(gs_mj6000c_device)
device_(gs_mj700v2c_device)
device_(gs_mj8000c_device)
device_(gs_ml600_device)
device_(gs_necp6_device)
device_(gs_npdl_device)
#endif
#ifdef device2_
device2_(gs_oce9050_device)
#endif
#ifdef device_
device_(gs_oki182_device)
device_(gs_oki4w_device)
device_(gs_okiibm_device)
device_(gs_oprp_device)
device_(gs_opvp_device)
device_(gs_paintjet_device)
device_(gs_pcl3_device)
device_(gs_photoex_device)
device_(gs_picty180_device)
device_(gs_pj_device)
device_(gs_pjetxl_device)
device_(gs_pjxl_device)
device_(gs_pjxl300_device)
device_(gs_pr1000_device)
device_(gs_pr1000_4_device)
device_(gs_pr150_device)
device_(gs_pr201_device)
#endif
#ifdef device2_
device2_(gs_pxlcolor_device)
device2_(gs_pxlmono_device)
#endif
#ifdef device_
device_(gs_r4081_device)
device_(gs_rinkj_device)
device_(gs_rpdl_device)
device_(gs_samsunggdi_device)
device_(gs_sj48_device)
device_(gs_st800_device)
device_(gs_stcolor_device)
device_(gs_t4693d2_device)
device_(gs_t4693d4_device)
device_(gs_t4693d8_device)
device_(gs_tek4696_device)
device_(gs_uniprint_device)
device_(gs_xes_device)
#endif
#ifdef device2_
device2_(gs_appleraster_device)
device2_(gs_cups_device)
device2_(gs_pwgraster_device)
device2_(gs_urf_device)
#endif
#ifdef device_
device_(gs_ijs_device)
#endif
#ifdef device2_
device2_(gs_png16_device)
device2_(gs_png16m_device)
device2_(gs_png16malpha_device)
device2_(gs_png256_device)
device2_(gs_png48_device)
device2_(gs_pngalpha_device)
device2_(gs_pnggray_device)
device2_(gs_pngmono_device)
device2_(gs_pngmonod_device)
device2_(gs_ocr_device)
device2_(gs_hocr_device)
device2_(gs_pdfocr8_device)
device2_(gs_pdfocr24_device)
device2_(gs_pdfocr32_device)
device2_(gs_nullpage_device)
#endif
#ifdef io_device_
io_device_(gs_iodev_pipe)
#endif
#ifdef image_type_
image_type_(3,gs_image_type_3)
image_type_(4,gs_image_type_4)
image_type_(3x,gs_image_type_3x)
#endif
#ifdef init_
init_(gs_gxclfile_init)
init_(gs_gxclmem_init)
#endif
#ifdef io_device_
io_device_(gs_iodev_ram)
#endif
#ifdef init_
init_(gs_gshtscr_init)
#endif
#ifdef io_device_
io_device_(gs_iodev_stdin)
io_device_(gs_iodev_stdout)
io_device_(gs_iodev_stderr)
io_device_(gs_iodev_rom)
#endif
#ifdef fapi_
fapi_(gs_fapi_ft_init)
#endif
#ifdef init_
init_(gs_fapi_init)
#endif
#ifdef image_type_
image_type_(1,gs_image_type_1)
image_type_(mask1,gs_image_type_mask1)
#endif
#ifdef image_class_
image_class_(gs_image_class_0_interpolate)
image_class_(gs_image_class_1_simple)
image_class_(gs_image_class_2_fracs)
image_class_(gs_image_class_3_mono)
image_class_(gs_image_class_4_color)
#endif

