/* THIS IS AUTO-GENERATED CODE, DO NOT EDIT. */

#include "docx_template.h"

const docx_template_item_t docx_template_items[] =
{
    {
        "[Content_Types].xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<Types xmlns=\"http://schemas.openxmlformats.org/package/2006/content-types\">"
                "<Default Extension=\"rels\" ContentType=\"application/vnd.openxmlformats-package.relationships+xml\"/>"
                "<Default Extension=\"xml\" ContentType=\"application/xml\"/>"
                "<Override PartName=\"/word/document.xml\" ContentType=\"application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml\"/>"
                "<Override PartName=\"/word/styles.xml\" ContentType=\"application/vnd.openxmlformats-officedocument.wordprocessingml.styles+xml\"/>"
                "<Override PartName=\"/word/settings.xml\" ContentType=\"application/vnd.openxmlformats-officedocument.wordprocessingml.settings+xml\"/>"
                "<Override PartName=\"/word/webSettings.xml\" ContentType=\"application/vnd.openxmlformats-officedocument.wordprocessingml.webSettings+xml\"/>"
                "<Override PartName=\"/word/fontTable.xml\" ContentType=\"application/vnd.openxmlformats-officedocument.wordprocessingml.fontTable+xml\"/>"
                "<Override PartName=\"/word/theme/theme1.xml\" ContentType=\"application/vnd.openxmlformats-officedocument.theme+xml\"/>"
                "<Override PartName=\"/docProps/core.xml\" ContentType=\"application/vnd.openxmlformats-package.core-properties+xml\"/>"
                "<Override PartName=\"/docProps/app.xml\" ContentType=\"application/vnd.openxmlformats-officedocument.extended-properties+xml\"/></Types>"
    },

    {
        "_rels/.rels",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<Relationships xmlns=\"http://schemas.openxmlformats.org/package/2006/relationships\">"
                "<Relationship Id=\"rId3\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties\" Target=\"docProps/app.xml\"/>"
                "<Relationship Id=\"rId2\" Type=\"http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties\" Target=\"docProps/core.xml\"/>"
                "<Relationship Id=\"rId1\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument\" Target=\"word/document.xml\"/></Relationships>"
    },

    {
        "docProps/app.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<Properties xmlns=\"http://schemas.openxmlformats.org/officeDocument/2006/extended-properties\" xmlns:vt=\"http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes\">"
                "<Template>Normal.dotm</Template>"
                "<TotalTime>3</TotalTime>"
                "<Pages>1</Pages>"
                "<Words>2</Words>"
                "<Characters>18</Characters>"
                "<Application>Microsoft Office Word</Application>"
                "<DocSecurity>0</DocSecurity>"
                "<Lines>1</Lines>"
                "<Paragraphs>1</Paragraphs>"
                "<ScaleCrop>false</ScaleCrop>"
                "<Company></Company>"
                "<LinksUpToDate>false</LinksUpToDate>"
                "<CharactersWithSpaces>19</CharactersWithSpaces>"
                "<SharedDoc>false</SharedDoc>"
                "<HyperlinksChanged>false</HyperlinksChanged>"
                "<AppVersion>16.0000</AppVersion></Properties>"
    },

    {
        "docProps/core.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<cp:coreProperties xmlns:cp=\"http://schemas.openxmlformats.org/package/2006/metadata/core-properties\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:dcterms=\"http://purl.org/dc/terms/\" xmlns:dcmitype=\"http://purl.org/dc/dcmitype/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">"
                "<dc:title></dc:title>"
                "<dc:subject></dc:subject>"
                "<dc:creator></dc:creator>"
                "<cp:keywords></cp:keywords>"
                "<dc:description></dc:description>"
                "<cp:lastModifiedBy></cp:lastModifiedBy>"
                "<cp:revision>1</cp:revision>"
                "<dcterms:created xsi:type=\"dcterms:W3CDTF\">2020-09-25T17:04:00Z</dcterms:created>"
                "<dcterms:modified xsi:type=\"dcterms:W3CDTF\">2020-09-25T17:07:00Z</dcterms:modified></cp:coreProperties>"
    },

    {
        "word/document.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<w:document xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex wp14\">"
                "<w:body>"
                "<w:p w14:paraId=\"7C58A6F1\" w14:textId=\"3E2CAE3F\" w:rsidR=\"00610D78\" w:rsidRDefault=\"007F4427\">"
                "<w:r>"
                "<w:t>Hello world</w:t></w:r></w:p>"
                "<w:p w14:paraId=\"53256C58\" w14:textId=\"13022069\" w:rsidR=\"007F4427\" w:rsidRDefault=\"007F4427\">"
                "<w:r>"
                "<w:rPr>"
                "<w:noProof/></w:rPr>"
                "<mc:AlternateContent>"
                "<mc:Choice Requires=\"wps\">"
                "<w:drawing>"
                "<wp:anchor distT=\"0\" distB=\"0\" distL=\"114300\" distR=\"114300\" simplePos=\"0\" relativeHeight=\"251659264\" behindDoc=\"0\" locked=\"0\" layoutInCell=\"1\" allowOverlap=\"1\" wp14:anchorId=\"53A210D1\" wp14:editId=\"2B7E8016\">"
                "<wp:simplePos x=\"0\" y=\"0\"/>"
                "<wp:positionH relativeFrom=\"column\">"
                "<wp:posOffset>904875</wp:posOffset></wp:positionH>"
                "<wp:positionV relativeFrom=\"paragraph\">"
                "<wp:posOffset>619125</wp:posOffset></wp:positionV>"
                "<wp:extent cx=\"3228975\" cy=\"2286000\"/>"
                "<wp:effectExtent l=\"381000\" t=\"723900\" r=\"371475\" b=\"723900\"/>"
                "<wp:wrapNone/>"
                "<wp:docPr id=\"1\" name=\"Text Box 1\"/>"
                "<wp:cNvGraphicFramePr/>"
                "<a:graphic xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\">"
                "<a:graphicData uri=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\">"
                "<wps:wsp>"
                "<wps:cNvSpPr txBox=\"1\"/>"
                "<wps:spPr>"
                "<a:xfrm rot=\"19547867\">"
                "<a:off x=\"0\" y=\"0\"/>"
                "<a:ext cx=\"3228975\" cy=\"2286000\"/></a:xfrm>"
                "<a:prstGeom prst=\"rect\">"
                "<a:avLst/></a:prstGeom>"
                "<a:solidFill>"
                "<a:schemeClr val=\"lt1\"/></a:solidFill>"
                "<a:ln w=\"6350\">"
                "<a:solidFill>"
                "<a:prstClr val=\"black\"/></a:solidFill></a:ln></wps:spPr>"
                "<wps:txbx>"
                "<w:txbxContent>"
                "<w:p w14:paraId=\"31597E69\" w14:textId=\"2903B1F1\" w:rsidR=\"007F4427\" w:rsidRDefault=\"007F4427\">"
                "<w:r>"
                "<w:t>Hello. Qwerty. World</w:t></w:r></w:p>"
                "<w:p w14:paraId=\"0BD8A985\" w14:textId=\"1BFB8248\" w:rsidR=\"007F4427\" w:rsidRDefault=\"007F4427\">"
                "<w:proofErr w:type=\"spellStart\"/>"
                "<w:r>"
                "<w:t>mupdf</w:t></w:r>"
                "<w:proofErr w:type=\"spellEnd\"/></w:p></w:txbxContent></wps:txbx>"
                "<wps:bodyPr rot=\"0\" spcFirstLastPara=\"0\" vertOverflow=\"overflow\" horzOverflow=\"overflow\" vert=\"horz\" wrap=\"square\" lIns=\"91440\" tIns=\"45720\" rIns=\"91440\" bIns=\"45720\" numCol=\"1\" spcCol=\"0\" rtlCol=\"0\" fromWordArt=\"0\" anchor=\"t\" anchorCtr=\"0\" forceAA=\"0\" compatLnSpc=\"1\">"
                "<a:prstTxWarp prst=\"textNoShape\">"
                "<a:avLst/></a:prstTxWarp>"
                "<a:noAutofit/></wps:bodyPr></wps:wsp></a:graphicData></a:graphic></wp:anchor></w:drawing></mc:Choice>"
                "<mc:Fallback>"
                "<w:pict>"
                "<v:shapetype w14:anchorId=\"53A210D1\" id=\"_x0000_t202\" coordsize=\"21600,21600\" o:spt=\"202\" path=\"m,l,21600r21600,l21600,xe\">"
                "<v:stroke joinstyle=\"miter\"/>"
                "<v:path gradientshapeok=\"t\" o:connecttype=\"rect\"/></v:shapetype>"
                "<v:shape id=\"Text Box 1\" o:spid=\"_x0000_s1026\" type=\"#_x0000_t202\" style=\"position:absolute;margin-left:71.25pt;margin-top:48.75pt;width:254.25pt;height:180pt;rotation:-2241476fd;z-index:251659264;visibility:visible;mso-wrap-style:square;mso-wrap-distance-left:9pt;mso-wrap-distance-top:0;mso-wrap-distance-right:9pt;mso-wrap-distance-bottom:0;mso-position-horizontal:absolute;mso-position-horizontal-relative:text;mso-position-vertical:absolute;mso-position-vertical-relative:text;v-text-anchor:top\" o:gfxdata=\"UEsDBBQABgAIAAAAIQC2gziS/gAAAOEBAAATAAAAW0NvbnRlbnRfVHlwZXNdLnhtbJSRQU7DMBBF&#xA;90jcwfIWJU67QAgl6YK0S0CoHGBkTxKLZGx5TGhvj5O2G0SRWNoz/78nu9wcxkFMGNg6quQqL6RA&#xA;0s5Y6ir5vt9lD1JwBDIwOMJKHpHlpr69KfdHjyxSmriSfYz+USnWPY7AufNIadK6MEJMx9ApD/oD&#xA;OlTrorhX2lFEilmcO2RdNtjC5xDF9pCuTyYBB5bi6bQ4syoJ3g9WQ0ymaiLzg5KdCXlKLjvcW893&#xA;SUOqXwnz5DrgnHtJTxOsQfEKIT7DmDSUCaxw7Rqn8787ZsmRM9e2VmPeBN4uqYvTtW7jvijg9N/y&#xA;JsXecLq0q+WD6m8AAAD//wMAUEsDBBQABgAIAAAAIQA4/SH/1gAAAJQBAAALAAAAX3JlbHMvLnJl&#xA;bHOkkMFqwzAMhu+DvYPRfXGawxijTi+j0GvpHsDYimMaW0Yy2fr2M4PBMnrbUb/Q94l/f/hMi1qR&#xA;JVI2sOt6UJgd+ZiDgffL8ekFlFSbvV0oo4EbChzGx4f9GRdb25HMsYhqlCwG5lrLq9biZkxWOiqY&#xA;22YiTra2kYMu1l1tQD30/bPm3wwYN0x18gb45AdQl1tp5j/sFB2T0FQ7R0nTNEV3j6o9feQzro1i&#xA;OWA14Fm+Q8a1a8+Bvu/d/dMb2JY5uiPbhG/ktn4cqGU/er3pcvwCAAD//wMAUEsDBBQABgAIAAAA&#xA;IQDQg5pQVgIAALEEAAAOAAAAZHJzL2Uyb0RvYy54bWysVE1v2zAMvQ/YfxB0X+2k+WiDOEXWosOA&#xA;oi3QDj0rstwYk0VNUmJ3v35PipMl3U7DLgJFPj+Rj6TnV12j2VY5X5Mp+OAs50wZSWVtXgv+7fn2&#xA;0wVnPghTCk1GFfxNeX61+Phh3tqZGtKadKkcA4nxs9YWfB2CnWWZl2vVCH9GVhkEK3KNCLi616x0&#xA;ogV7o7Nhnk+yllxpHUnlPbw3uyBfJP6qUjI8VJVXgemCI7eQTpfOVTyzxVzMXp2w61r2aYh/yKIR&#xA;tcGjB6obEQTbuPoPqqaWjjxV4UxSk1FV1VKlGlDNIH9XzdNaWJVqgTjeHmTy/49W3m8fHatL9I4z&#xA;Ixq06Fl1gX2mjg2iOq31M4CeLGChgzsie7+HMxbdVa5hjiDu4HI8ml5MpkkLVMcAh+xvB6kjt4Tz&#xA;fDi8uJyOOZOIwZ7keWpGtmOLrNb58EVRw6JRcIdeJlqxvfMBGQC6h0S4J12Xt7XW6RLnR11rx7YC&#xA;ndch5YwvTlDasLbgk/NxnohPYpH68P1KC/k9Vn3KgJs2cEaNdlpEK3SrrhdoReUbdEvSQAZv5W0N&#xA;3jvhw6NwGDQ4sTzhAUelCclQb3G2Jvfzb/6IR/8R5azF4Bbc/9gIpzjTXw0m43IwGsVJT5fReDrE&#xA;xR1HVscRs2muCQqh+8gumREf9N6sHDUv2LFlfBUhYSTeLnjYm9dht07YUamWywTCbFsR7syTlZF6&#xA;383n7kU42/czYBTuaT/iYvaurTts/NLQchOoqlPPo8A7VXvdsRepLf0Ox8U7vifU7z/N4hcAAAD/&#xA;/wMAUEsDBBQABgAIAAAAIQBh17L63wAAAAoBAAAPAAAAZHJzL2Rvd25yZXYueG1sTI9BT4NAEIXv&#xA;Jv6HzZh4s0ubgpayNIboSW3Syg9Y2BGI7CyyS0v99Y4nPU3ezMub72W72fbihKPvHClYLiIQSLUz&#xA;HTUKyvfnuwcQPmgyuneECi7oYZdfX2U6Ne5MBzwdQyM4hHyqFbQhDKmUvm7Rar9wAxLfPtxodWA5&#xA;NtKM+szhtperKEqk1R3xh1YPWLRYfx4nq8APVfz9VQxPb+WUNC+vZbGPDhelbm/mxy2IgHP4M8Mv&#xA;PqNDzkyVm8h40bNer2K2Ktjc82RDEi+5XKVgHfNG5pn8XyH/AQAA//8DAFBLAQItABQABgAIAAAA&#xA;IQC2gziS/gAAAOEBAAATAAAAAAAAAAAAAAAAAAAAAABbQ29udGVudF9UeXBlc10ueG1sUEsBAi0A&#xA;FAAGAAgAAAAhADj9If/WAAAAlAEAAAsAAAAAAAAAAAAAAAAALwEAAF9yZWxzLy5yZWxzUEsBAi0A&#xA;FAAGAAgAAAAhANCDmlBWAgAAsQQAAA4AAAAAAAAAAAAAAAAALgIAAGRycy9lMm9Eb2MueG1sUEsB&#xA;Ai0AFAAGAAgAAAAhAGHXsvrfAAAACgEAAA8AAAAAAAAAAAAAAAAAsAQAAGRycy9kb3ducmV2Lnht&#xA;bFBLBQYAAAAABAAEAPMAAAC8BQAAAAA=&#xA;\" fillcolor=\"white [3201]\" strokeweight=\".5pt\">"
                "<v:textbox>"
                "<w:txbxContent>"
                "<w:p w14:paraId=\"31597E69\" w14:textId=\"2903B1F1\" w:rsidR=\"007F4427\" w:rsidRDefault=\"007F4427\">"
                "<w:r>"
                "<w:t>Hello. Qwerty. World</w:t></w:r></w:p>"
                "<w:p w14:paraId=\"0BD8A985\" w14:textId=\"1BFB8248\" w:rsidR=\"007F4427\" w:rsidRDefault=\"007F4427\">"
                "<w:proofErr w:type=\"spellStart\"/>"
                "<w:r>"
                "<w:t>mupdf</w:t></w:r>"
                "<w:proofErr w:type=\"spellEnd\"/></w:p></w:txbxContent></v:textbox></v:shape></w:pict></mc:Fallback></mc:AlternateContent></w:r>"
                "<w:r>"
                "<w:t>qwerty</w:t></w:r></w:p>"
                "<w:sectPr w:rsidR=\"007F4427\">"
                "<w:pgSz w:w=\"11906\" w:h=\"16838\"/>"
                "<w:pgMar w:top=\"1440\" w:right=\"1440\" w:bottom=\"1440\" w:left=\"1440\" w:header=\"708\" w:footer=\"708\" w:gutter=\"0\"/>"
                "<w:cols w:space=\"708\"/>"
                "<w:docGrid w:linePitch=\"360\"/></w:sectPr></w:body></w:document>"
    },

    {
        "word/fontTable.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<w:fonts xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex\">"
                "<w:font w:name=\"Calibri\">"
                "<w:panose1 w:val=\"020F0502020204030204\"/>"
                "<w:charset w:val=\"00\"/>"
                "<w:family w:val=\"swiss\"/>"
                "<w:pitch w:val=\"variable\"/>"
                "<w:sig w:usb0=\"E4002EFF\" w:usb1=\"C000247B\" w:usb2=\"00000009\" w:usb3=\"00000000\" w:csb0=\"000001FF\" w:csb1=\"00000000\"/></w:font>"
                "<w:font w:name=\"Times New Roman\">"
                "<w:panose1 w:val=\"02020603050405020304\"/>"
                "<w:charset w:val=\"00\"/>"
                "<w:family w:val=\"roman\"/>"
                "<w:pitch w:val=\"variable\"/>"
                "<w:sig w:usb0=\"E0002EFF\" w:usb1=\"C000785B\" w:usb2=\"00000009\" w:usb3=\"00000000\" w:csb0=\"000001FF\" w:csb1=\"00000000\"/></w:font>"
                "<w:font w:name=\"Calibri Light\">"
                "<w:panose1 w:val=\"020F0302020204030204\"/>"
                "<w:charset w:val=\"00\"/>"
                "<w:family w:val=\"swiss\"/>"
                "<w:pitch w:val=\"variable\"/>"
                "<w:sig w:usb0=\"E4002EFF\" w:usb1=\"C000247B\" w:usb2=\"00000009\" w:usb3=\"00000000\" w:csb0=\"000001FF\" w:csb1=\"00000000\"/></w:font></w:fonts>"
    },

    {
        "word/settings.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<w:settings xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:sl=\"http://schemas.openxmlformats.org/schemaLibrary/2006/main\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex\">"
                "<w:zoom w:percent=\"100\"/>"
                "<w:proofState w:spelling=\"clean\" w:grammar=\"clean\"/>"
                "<w:defaultTabStop w:val=\"720\"/>"
                "<w:characterSpacingControl w:val=\"doNotCompress\"/>"
                "<w:compat>"
                "<w:compatSetting w:name=\"compatibilityMode\" w:uri=\"http://schemas.microsoft.com/office/word\" w:val=\"15\"/>"
                "<w:compatSetting w:name=\"overrideTableStyleFontSizeAndJustification\" w:uri=\"http://schemas.microsoft.com/office/word\" w:val=\"1\"/>"
                "<w:compatSetting w:name=\"enableOpenTypeFeatures\" w:uri=\"http://schemas.microsoft.com/office/word\" w:val=\"1\"/>"
                "<w:compatSetting w:name=\"doNotFlipMirrorIndents\" w:uri=\"http://schemas.microsoft.com/office/word\" w:val=\"1\"/>"
                "<w:compatSetting w:name=\"differentiateMultirowTableHeaders\" w:uri=\"http://schemas.microsoft.com/office/word\" w:val=\"1\"/>"
                "<w:compatSetting w:name=\"useWord2013TrackBottomHyphenation\" w:uri=\"http://schemas.microsoft.com/office/word\" w:val=\"0\"/></w:compat>"
                "<w:rsids>"
                "<w:rsidRoot w:val=\"007F4427\"/>"
                "<w:rsid w:val=\"00255448\"/>"
                "<w:rsid w:val=\"007F4427\"/></w:rsids>"
                "<m:mathPr>"
                "<m:mathFont m:val=\"Cambria Math\"/>"
                "<m:brkBin m:val=\"before\"/>"
                "<m:brkBinSub m:val=\"--\"/>"
                "<m:smallFrac m:val=\"0\"/>"
                "<m:dispDef/>"
                "<m:lMargin m:val=\"0\"/>"
                "<m:rMargin m:val=\"0\"/>"
                "<m:defJc m:val=\"centerGroup\"/>"
                "<m:wrapIndent m:val=\"1440\"/>"
                "<m:intLim m:val=\"subSup\"/>"
                "<m:naryLim m:val=\"undOvr\"/></m:mathPr>"
                "<w:themeFontLang w:val=\"en-GB\"/>"
                "<w:clrSchemeMapping w:bg1=\"light1\" w:t1=\"dark1\" w:bg2=\"light2\" w:t2=\"dark2\" w:accent1=\"accent1\" w:accent2=\"accent2\" w:accent3=\"accent3\" w:accent4=\"accent4\" w:accent5=\"accent5\" w:accent6=\"accent6\" w:hyperlink=\"hyperlink\" w:followedHyperlink=\"followedHyperlink\"/>"
                "<w:shapeDefaults>"
                "<o:shapedefaults v:ext=\"edit\" spidmax=\"1026\"/>"
                "<o:shapelayout v:ext=\"edit\">"
                "<o:idmap v:ext=\"edit\" data=\"1\"/></o:shapelayout></w:shapeDefaults>"
                "<w:decimalSymbol w:val=\".\"/>"
                "<w:listSeparator w:val=\",\"/>"
                "<w14:docId w14:val=\"32E52EF8\"/>"
                "<w15:chartTrackingRefBased/>"
                "<w15:docId w15:val=\"{A10F59F7-497D-44D4-A338-47719734E7A0}\"/></w:settings>"
    },

    {
        "word/styles.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<w:styles xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex\">"
                "<w:docDefaults>"
                "<w:rPrDefault>"
                "<w:rPr>"
                "<w:rFonts w:asciiTheme=\"minorHAnsi\" w:eastAsiaTheme=\"minorHAnsi\" w:hAnsiTheme=\"minorHAnsi\" w:cstheme=\"minorBidi\"/>"
                "<w:sz w:val=\"22\"/>"
                "<w:szCs w:val=\"22\"/>"
                "<w:lang w:val=\"en-GB\" w:eastAsia=\"en-US\" w:bidi=\"ar-SA\"/></w:rPr></w:rPrDefault>"
                "<w:pPrDefault>"
                "<w:pPr>"
                "<w:spacing w:after=\"160\" w:line=\"259\" w:lineRule=\"auto\"/></w:pPr></w:pPrDefault></w:docDefaults>"
                "<w:latentStyles w:defLockedState=\"0\" w:defUIPriority=\"99\" w:defSemiHidden=\"0\" w:defUnhideWhenUsed=\"0\" w:defQFormat=\"0\" w:count=\"376\">"
                "<w:lsdException w:name=\"Normal\" w:uiPriority=\"0\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 1\" w:uiPriority=\"9\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 2\" w:semiHidden=\"1\" w:uiPriority=\"9\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 3\" w:semiHidden=\"1\" w:uiPriority=\"9\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 4\" w:semiHidden=\"1\" w:uiPriority=\"9\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 5\" w:semiHidden=\"1\" w:uiPriority=\"9\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 6\" w:semiHidden=\"1\" w:uiPriority=\"9\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 7\" w:semiHidden=\"1\" w:uiPriority=\"9\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 8\" w:semiHidden=\"1\" w:uiPriority=\"9\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"heading 9\" w:semiHidden=\"1\" w:uiPriority=\"9\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"index 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index 5\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index 6\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index 7\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index 8\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index 9\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 1\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 2\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 3\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 4\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 5\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 6\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 7\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 8\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toc 9\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Normal Indent\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"footnote text\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"annotation text\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"header\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"footer\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"index heading\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"caption\" w:semiHidden=\"1\" w:uiPriority=\"35\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"table of figures\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"envelope address\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"envelope return\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"footnote reference\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"annotation reference\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"line number\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"page number\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"endnote reference\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"endnote text\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"table of authorities\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"macro\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"toa heading\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Bullet\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Number\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List 5\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Bullet 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Bullet 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Bullet 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Bullet 5\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Number 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Number 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Number 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Number 5\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Title\" w:uiPriority=\"10\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Closing\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Signature\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Default Paragraph Font\" w:semiHidden=\"1\" w:uiPriority=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Body Text\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Body Text Indent\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Continue\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Continue 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Continue 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Continue 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"List Continue 5\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Message Header\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Subtitle\" w:uiPriority=\"11\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Salutation\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Date\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Body Text First Indent\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Body Text First Indent 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Note Heading\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Body Text 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Body Text 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Body Text Indent 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Body Text Indent 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Block Text\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Hyperlink\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"FollowedHyperlink\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Strong\" w:uiPriority=\"22\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Emphasis\" w:uiPriority=\"20\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Document Map\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Plain Text\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"E-mail Signature\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Top of Form\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Bottom of Form\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Normal (Web)\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Acronym\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Address\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Cite\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Code\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Definition\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Keyboard\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Preformatted\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Sample\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Typewriter\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"HTML Variable\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Normal Table\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"annotation subject\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"No List\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Outline List 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Outline List 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Outline List 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Simple 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Simple 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Simple 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Classic 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Classic 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Classic 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Classic 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Colorful 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Colorful 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Colorful 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Columns 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Columns 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Columns 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Columns 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Columns 5\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid 5\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid 6\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid 7\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid 8\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table List 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table List 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table List 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table List 4\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table List 5\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table List 6\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table List 7\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table List 8\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table 3D effects 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table 3D effects 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table 3D effects 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Contemporary\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Elegant\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Professional\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Subtle 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Subtle 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Web 1\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Web 2\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Web 3\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Balloon Text\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Table Grid\" w:uiPriority=\"39\"/>"
                "<w:lsdException w:name=\"Table Theme\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Placeholder Text\" w:semiHidden=\"1\"/>"
                "<w:lsdException w:name=\"No Spacing\" w:uiPriority=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Light Shading\" w:uiPriority=\"60\"/>"
                "<w:lsdException w:name=\"Light List\" w:uiPriority=\"61\"/>"
                "<w:lsdException w:name=\"Light Grid\" w:uiPriority=\"62\"/>"
                "<w:lsdException w:name=\"Medium Shading 1\" w:uiPriority=\"63\"/>"
                "<w:lsdException w:name=\"Medium Shading 2\" w:uiPriority=\"64\"/>"
                "<w:lsdException w:name=\"Medium List 1\" w:uiPriority=\"65\"/>"
                "<w:lsdException w:name=\"Medium List 2\" w:uiPriority=\"66\"/>"
                "<w:lsdException w:name=\"Medium Grid 1\" w:uiPriority=\"67\"/>"
                "<w:lsdException w:name=\"Medium Grid 2\" w:uiPriority=\"68\"/>"
                "<w:lsdException w:name=\"Medium Grid 3\" w:uiPriority=\"69\"/>"
                "<w:lsdException w:name=\"Dark List\" w:uiPriority=\"70\"/>"
                "<w:lsdException w:name=\"Colorful Shading\" w:uiPriority=\"71\"/>"
                "<w:lsdException w:name=\"Colorful List\" w:uiPriority=\"72\"/>"
                "<w:lsdException w:name=\"Colorful Grid\" w:uiPriority=\"73\"/>"
                "<w:lsdException w:name=\"Light Shading Accent 1\" w:uiPriority=\"60\"/>"
                "<w:lsdException w:name=\"Light List Accent 1\" w:uiPriority=\"61\"/>"
                "<w:lsdException w:name=\"Light Grid Accent 1\" w:uiPriority=\"62\"/>"
                "<w:lsdException w:name=\"Medium Shading 1 Accent 1\" w:uiPriority=\"63\"/>"
                "<w:lsdException w:name=\"Medium Shading 2 Accent 1\" w:uiPriority=\"64\"/>"
                "<w:lsdException w:name=\"Medium List 1 Accent 1\" w:uiPriority=\"65\"/>"
                "<w:lsdException w:name=\"Revision\" w:semiHidden=\"1\"/>"
                "<w:lsdException w:name=\"List Paragraph\" w:uiPriority=\"34\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Quote\" w:uiPriority=\"29\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Intense Quote\" w:uiPriority=\"30\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Medium List 2 Accent 1\" w:uiPriority=\"66\"/>"
                "<w:lsdException w:name=\"Medium Grid 1 Accent 1\" w:uiPriority=\"67\"/>"
                "<w:lsdException w:name=\"Medium Grid 2 Accent 1\" w:uiPriority=\"68\"/>"
                "<w:lsdException w:name=\"Medium Grid 3 Accent 1\" w:uiPriority=\"69\"/>"
                "<w:lsdException w:name=\"Dark List Accent 1\" w:uiPriority=\"70\"/>"
                "<w:lsdException w:name=\"Colorful Shading Accent 1\" w:uiPriority=\"71\"/>"
                "<w:lsdException w:name=\"Colorful List Accent 1\" w:uiPriority=\"72\"/>"
                "<w:lsdException w:name=\"Colorful Grid Accent 1\" w:uiPriority=\"73\"/>"
                "<w:lsdException w:name=\"Light Shading Accent 2\" w:uiPriority=\"60\"/>"
                "<w:lsdException w:name=\"Light List Accent 2\" w:uiPriority=\"61\"/>"
                "<w:lsdException w:name=\"Light Grid Accent 2\" w:uiPriority=\"62\"/>"
                "<w:lsdException w:name=\"Medium Shading 1 Accent 2\" w:uiPriority=\"63\"/>"
                "<w:lsdException w:name=\"Medium Shading 2 Accent 2\" w:uiPriority=\"64\"/>"
                "<w:lsdException w:name=\"Medium List 1 Accent 2\" w:uiPriority=\"65\"/>"
                "<w:lsdException w:name=\"Medium List 2 Accent 2\" w:uiPriority=\"66\"/>"
                "<w:lsdException w:name=\"Medium Grid 1 Accent 2\" w:uiPriority=\"67\"/>"
                "<w:lsdException w:name=\"Medium Grid 2 Accent 2\" w:uiPriority=\"68\"/>"
                "<w:lsdException w:name=\"Medium Grid 3 Accent 2\" w:uiPriority=\"69\"/>"
                "<w:lsdException w:name=\"Dark List Accent 2\" w:uiPriority=\"70\"/>"
                "<w:lsdException w:name=\"Colorful Shading Accent 2\" w:uiPriority=\"71\"/>"
                "<w:lsdException w:name=\"Colorful List Accent 2\" w:uiPriority=\"72\"/>"
                "<w:lsdException w:name=\"Colorful Grid Accent 2\" w:uiPriority=\"73\"/>"
                "<w:lsdException w:name=\"Light Shading Accent 3\" w:uiPriority=\"60\"/>"
                "<w:lsdException w:name=\"Light List Accent 3\" w:uiPriority=\"61\"/>"
                "<w:lsdException w:name=\"Light Grid Accent 3\" w:uiPriority=\"62\"/>"
                "<w:lsdException w:name=\"Medium Shading 1 Accent 3\" w:uiPriority=\"63\"/>"
                "<w:lsdException w:name=\"Medium Shading 2 Accent 3\" w:uiPriority=\"64\"/>"
                "<w:lsdException w:name=\"Medium List 1 Accent 3\" w:uiPriority=\"65\"/>"
                "<w:lsdException w:name=\"Medium List 2 Accent 3\" w:uiPriority=\"66\"/>"
                "<w:lsdException w:name=\"Medium Grid 1 Accent 3\" w:uiPriority=\"67\"/>"
                "<w:lsdException w:name=\"Medium Grid 2 Accent 3\" w:uiPriority=\"68\"/>"
                "<w:lsdException w:name=\"Medium Grid 3 Accent 3\" w:uiPriority=\"69\"/>"
                "<w:lsdException w:name=\"Dark List Accent 3\" w:uiPriority=\"70\"/>"
                "<w:lsdException w:name=\"Colorful Shading Accent 3\" w:uiPriority=\"71\"/>"
                "<w:lsdException w:name=\"Colorful List Accent 3\" w:uiPriority=\"72\"/>"
                "<w:lsdException w:name=\"Colorful Grid Accent 3\" w:uiPriority=\"73\"/>"
                "<w:lsdException w:name=\"Light Shading Accent 4\" w:uiPriority=\"60\"/>"
                "<w:lsdException w:name=\"Light List Accent 4\" w:uiPriority=\"61\"/>"
                "<w:lsdException w:name=\"Light Grid Accent 4\" w:uiPriority=\"62\"/>"
                "<w:lsdException w:name=\"Medium Shading 1 Accent 4\" w:uiPriority=\"63\"/>"
                "<w:lsdException w:name=\"Medium Shading 2 Accent 4\" w:uiPriority=\"64\"/>"
                "<w:lsdException w:name=\"Medium List 1 Accent 4\" w:uiPriority=\"65\"/>"
                "<w:lsdException w:name=\"Medium List 2 Accent 4\" w:uiPriority=\"66\"/>"
                "<w:lsdException w:name=\"Medium Grid 1 Accent 4\" w:uiPriority=\"67\"/>"
                "<w:lsdException w:name=\"Medium Grid 2 Accent 4\" w:uiPriority=\"68\"/>"
                "<w:lsdException w:name=\"Medium Grid 3 Accent 4\" w:uiPriority=\"69\"/>"
                "<w:lsdException w:name=\"Dark List Accent 4\" w:uiPriority=\"70\"/>"
                "<w:lsdException w:name=\"Colorful Shading Accent 4\" w:uiPriority=\"71\"/>"
                "<w:lsdException w:name=\"Colorful List Accent 4\" w:uiPriority=\"72\"/>"
                "<w:lsdException w:name=\"Colorful Grid Accent 4\" w:uiPriority=\"73\"/>"
                "<w:lsdException w:name=\"Light Shading Accent 5\" w:uiPriority=\"60\"/>"
                "<w:lsdException w:name=\"Light List Accent 5\" w:uiPriority=\"61\"/>"
                "<w:lsdException w:name=\"Light Grid Accent 5\" w:uiPriority=\"62\"/>"
                "<w:lsdException w:name=\"Medium Shading 1 Accent 5\" w:uiPriority=\"63\"/>"
                "<w:lsdException w:name=\"Medium Shading 2 Accent 5\" w:uiPriority=\"64\"/>"
                "<w:lsdException w:name=\"Medium List 1 Accent 5\" w:uiPriority=\"65\"/>"
                "<w:lsdException w:name=\"Medium List 2 Accent 5\" w:uiPriority=\"66\"/>"
                "<w:lsdException w:name=\"Medium Grid 1 Accent 5\" w:uiPriority=\"67\"/>"
                "<w:lsdException w:name=\"Medium Grid 2 Accent 5\" w:uiPriority=\"68\"/>"
                "<w:lsdException w:name=\"Medium Grid 3 Accent 5\" w:uiPriority=\"69\"/>"
                "<w:lsdException w:name=\"Dark List Accent 5\" w:uiPriority=\"70\"/>"
                "<w:lsdException w:name=\"Colorful Shading Accent 5\" w:uiPriority=\"71\"/>"
                "<w:lsdException w:name=\"Colorful List Accent 5\" w:uiPriority=\"72\"/>"
                "<w:lsdException w:name=\"Colorful Grid Accent 5\" w:uiPriority=\"73\"/>"
                "<w:lsdException w:name=\"Light Shading Accent 6\" w:uiPriority=\"60\"/>"
                "<w:lsdException w:name=\"Light List Accent 6\" w:uiPriority=\"61\"/>"
                "<w:lsdException w:name=\"Light Grid Accent 6\" w:uiPriority=\"62\"/>"
                "<w:lsdException w:name=\"Medium Shading 1 Accent 6\" w:uiPriority=\"63\"/>"
                "<w:lsdException w:name=\"Medium Shading 2 Accent 6\" w:uiPriority=\"64\"/>"
                "<w:lsdException w:name=\"Medium List 1 Accent 6\" w:uiPriority=\"65\"/>"
                "<w:lsdException w:name=\"Medium List 2 Accent 6\" w:uiPriority=\"66\"/>"
                "<w:lsdException w:name=\"Medium Grid 1 Accent 6\" w:uiPriority=\"67\"/>"
                "<w:lsdException w:name=\"Medium Grid 2 Accent 6\" w:uiPriority=\"68\"/>"
                "<w:lsdException w:name=\"Medium Grid 3 Accent 6\" w:uiPriority=\"69\"/>"
                "<w:lsdException w:name=\"Dark List Accent 6\" w:uiPriority=\"70\"/>"
                "<w:lsdException w:name=\"Colorful Shading Accent 6\" w:uiPriority=\"71\"/>"
                "<w:lsdException w:name=\"Colorful List Accent 6\" w:uiPriority=\"72\"/>"
                "<w:lsdException w:name=\"Colorful Grid Accent 6\" w:uiPriority=\"73\"/>"
                "<w:lsdException w:name=\"Subtle Emphasis\" w:uiPriority=\"19\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Intense Emphasis\" w:uiPriority=\"21\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Subtle Reference\" w:uiPriority=\"31\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Intense Reference\" w:uiPriority=\"32\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Book Title\" w:uiPriority=\"33\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Bibliography\" w:semiHidden=\"1\" w:uiPriority=\"37\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"TOC Heading\" w:semiHidden=\"1\" w:uiPriority=\"39\" w:unhideWhenUsed=\"1\" w:qFormat=\"1\"/>"
                "<w:lsdException w:name=\"Plain Table 1\" w:uiPriority=\"41\"/>"
                "<w:lsdException w:name=\"Plain Table 2\" w:uiPriority=\"42\"/>"
                "<w:lsdException w:name=\"Plain Table 3\" w:uiPriority=\"43\"/>"
                "<w:lsdException w:name=\"Plain Table 4\" w:uiPriority=\"44\"/>"
                "<w:lsdException w:name=\"Plain Table 5\" w:uiPriority=\"45\"/>"
                "<w:lsdException w:name=\"Grid Table Light\" w:uiPriority=\"40\"/>"
                "<w:lsdException w:name=\"Grid Table 1 Light\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"Grid Table 2\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"Grid Table 3\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"Grid Table 4\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"Grid Table 5 Dark\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"Grid Table 6 Colorful\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"Grid Table 7 Colorful\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"Grid Table 1 Light Accent 1\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"Grid Table 2 Accent 1\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"Grid Table 3 Accent 1\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"Grid Table 4 Accent 1\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"Grid Table 5 Dark Accent 1\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"Grid Table 6 Colorful Accent 1\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"Grid Table 7 Colorful Accent 1\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"Grid Table 1 Light Accent 2\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"Grid Table 2 Accent 2\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"Grid Table 3 Accent 2\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"Grid Table 4 Accent 2\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"Grid Table 5 Dark Accent 2\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"Grid Table 6 Colorful Accent 2\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"Grid Table 7 Colorful Accent 2\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"Grid Table 1 Light Accent 3\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"Grid Table 2 Accent 3\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"Grid Table 3 Accent 3\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"Grid Table 4 Accent 3\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"Grid Table 5 Dark Accent 3\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"Grid Table 6 Colorful Accent 3\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"Grid Table 7 Colorful Accent 3\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"Grid Table 1 Light Accent 4\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"Grid Table 2 Accent 4\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"Grid Table 3 Accent 4\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"Grid Table 4 Accent 4\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"Grid Table 5 Dark Accent 4\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"Grid Table 6 Colorful Accent 4\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"Grid Table 7 Colorful Accent 4\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"Grid Table 1 Light Accent 5\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"Grid Table 2 Accent 5\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"Grid Table 3 Accent 5\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"Grid Table 4 Accent 5\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"Grid Table 5 Dark Accent 5\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"Grid Table 6 Colorful Accent 5\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"Grid Table 7 Colorful Accent 5\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"Grid Table 1 Light Accent 6\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"Grid Table 2 Accent 6\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"Grid Table 3 Accent 6\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"Grid Table 4 Accent 6\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"Grid Table 5 Dark Accent 6\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"Grid Table 6 Colorful Accent 6\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"Grid Table 7 Colorful Accent 6\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"List Table 1 Light\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"List Table 2\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"List Table 3\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"List Table 4\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"List Table 5 Dark\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"List Table 6 Colorful\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"List Table 7 Colorful\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"List Table 1 Light Accent 1\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"List Table 2 Accent 1\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"List Table 3 Accent 1\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"List Table 4 Accent 1\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"List Table 5 Dark Accent 1\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"List Table 6 Colorful Accent 1\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"List Table 7 Colorful Accent 1\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"List Table 1 Light Accent 2\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"List Table 2 Accent 2\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"List Table 3 Accent 2\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"List Table 4 Accent 2\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"List Table 5 Dark Accent 2\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"List Table 6 Colorful Accent 2\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"List Table 7 Colorful Accent 2\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"List Table 1 Light Accent 3\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"List Table 2 Accent 3\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"List Table 3 Accent 3\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"List Table 4 Accent 3\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"List Table 5 Dark Accent 3\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"List Table 6 Colorful Accent 3\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"List Table 7 Colorful Accent 3\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"List Table 1 Light Accent 4\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"List Table 2 Accent 4\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"List Table 3 Accent 4\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"List Table 4 Accent 4\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"List Table 5 Dark Accent 4\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"List Table 6 Colorful Accent 4\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"List Table 7 Colorful Accent 4\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"List Table 1 Light Accent 5\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"List Table 2 Accent 5\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"List Table 3 Accent 5\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"List Table 4 Accent 5\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"List Table 5 Dark Accent 5\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"List Table 6 Colorful Accent 5\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"List Table 7 Colorful Accent 5\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"List Table 1 Light Accent 6\" w:uiPriority=\"46\"/>"
                "<w:lsdException w:name=\"List Table 2 Accent 6\" w:uiPriority=\"47\"/>"
                "<w:lsdException w:name=\"List Table 3 Accent 6\" w:uiPriority=\"48\"/>"
                "<w:lsdException w:name=\"List Table 4 Accent 6\" w:uiPriority=\"49\"/>"
                "<w:lsdException w:name=\"List Table 5 Dark Accent 6\" w:uiPriority=\"50\"/>"
                "<w:lsdException w:name=\"List Table 6 Colorful Accent 6\" w:uiPriority=\"51\"/>"
                "<w:lsdException w:name=\"List Table 7 Colorful Accent 6\" w:uiPriority=\"52\"/>"
                "<w:lsdException w:name=\"Mention\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Smart Hyperlink\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Hashtag\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Unresolved Mention\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/>"
                "<w:lsdException w:name=\"Smart Link\" w:semiHidden=\"1\" w:unhideWhenUsed=\"1\"/></w:latentStyles>"
                "<w:style w:type=\"paragraph\" w:default=\"1\" w:styleId=\"Normal\">"
                "<w:name w:val=\"Normal\"/>"
                "<w:qFormat/></w:style>"
                "<w:style w:type=\"character\" w:default=\"1\" w:styleId=\"DefaultParagraphFont\">"
                "<w:name w:val=\"Default Paragraph Font\"/>"
                "<w:uiPriority w:val=\"1\"/>"
                "<w:semiHidden/>"
                "<w:unhideWhenUsed/></w:style>"
                "<w:style w:type=\"table\" w:default=\"1\" w:styleId=\"TableNormal\">"
                "<w:name w:val=\"Normal Table\"/>"
                "<w:uiPriority w:val=\"99\"/>"
                "<w:semiHidden/>"
                "<w:unhideWhenUsed/>"
                "<w:tblPr>"
                "<w:tblInd w:w=\"0\" w:type=\"dxa\"/>"
                "<w:tblCellMar>"
                "<w:top w:w=\"0\" w:type=\"dxa\"/>"
                "<w:left w:w=\"108\" w:type=\"dxa\"/>"
                "<w:bottom w:w=\"0\" w:type=\"dxa\"/>"
                "<w:right w:w=\"108\" w:type=\"dxa\"/></w:tblCellMar></w:tblPr></w:style>"
                "<w:style w:type=\"numbering\" w:default=\"1\" w:styleId=\"NoList\">"
                "<w:name w:val=\"No List\"/>"
                "<w:uiPriority w:val=\"99\"/>"
                "<w:semiHidden/>"
                "<w:unhideWhenUsed/></w:style></w:styles>"
    },

    {
        "word/webSettings.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<w:webSettings xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex\">"
                "<w:optimizeForBrowser/>"
                "<w:allowPNG/></w:webSettings>"
    },

    {
        "word/_rels/document.xml.rels",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<Relationships xmlns=\"http://schemas.openxmlformats.org/package/2006/relationships\">"
                "<Relationship Id=\"rId3\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/webSettings\" Target=\"webSettings.xml\"/>"
                "<Relationship Id=\"rId2\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/settings\" Target=\"settings.xml\"/>"
                "<Relationship Id=\"rId1\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/styles\" Target=\"styles.xml\"/>"
                "<Relationship Id=\"rId5\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme\" Target=\"theme/theme1.xml\"/>"
                "<Relationship Id=\"rId4\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/fontTable\" Target=\"fontTable.xml\"/></Relationships>"
    },

    {
        "word/theme/theme1.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\r\n"
                ""
                "<a:theme xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" name=\"Office Theme\">"
                "<a:themeElements>"
                "<a:clrScheme name=\"Office\">"
                "<a:dk1>"
                "<a:sysClr val=\"windowText\" lastClr=\"000000\"/></a:dk1>"
                "<a:lt1>"
                "<a:sysClr val=\"window\" lastClr=\"FFFFFF\"/></a:lt1>"
                "<a:dk2>"
                "<a:srgbClr val=\"44546A\"/></a:dk2>"
                "<a:lt2>"
                "<a:srgbClr val=\"E7E6E6\"/></a:lt2>"
                "<a:accent1>"
                "<a:srgbClr val=\"4472C4\"/></a:accent1>"
                "<a:accent2>"
                "<a:srgbClr val=\"ED7D31\"/></a:accent2>"
                "<a:accent3>"
                "<a:srgbClr val=\"A5A5A5\"/></a:accent3>"
                "<a:accent4>"
                "<a:srgbClr val=\"FFC000\"/></a:accent4>"
                "<a:accent5>"
                "<a:srgbClr val=\"5B9BD5\"/></a:accent5>"
                "<a:accent6>"
                "<a:srgbClr val=\"70AD47\"/></a:accent6>"
                "<a:hlink>"
                "<a:srgbClr val=\"0563C1\"/></a:hlink>"
                "<a:folHlink>"
                "<a:srgbClr val=\"954F72\"/></a:folHlink></a:clrScheme>"
                "<a:fontScheme name=\"Office\">"
                "<a:majorFont>"
                "<a:latin typeface=\"Calibri Light\" panose=\"020F0302020204030204\"/>"
                "<a:ea typeface=\"\"/>"
                "<a:cs typeface=\"\"/>"
                "<a:font script=\"Jpan\" typeface=\"\xe6\xb8\xb8\xe3\x82\xb4\xe3\x82\xb7\xe3\x83\x83\xe3\x82\xaf Light\"/>"
                "<a:font script=\"Hang\" typeface=\"\xeb\xa7\x91\xec\x9d\x80 \xea\xb3\xa0\xeb\x94\x95\"/>"
                "<a:font script=\"Hans\" typeface=\"\xe7\xad\x89\xe7\xba\xbf Light\"/>"
                "<a:font script=\"Hant\" typeface=\"\xe6\x96\xb0\xe7\xb4\xb0\xe6\x98\x8e\xe9\xab\x94\"/>"
                "<a:font script=\"Arab\" typeface=\"Times New Roman\"/>"
                "<a:font script=\"Hebr\" typeface=\"Times New Roman\"/>"
                "<a:font script=\"Thai\" typeface=\"Angsana New\"/>"
                "<a:font script=\"Ethi\" typeface=\"Nyala\"/>"
                "<a:font script=\"Beng\" typeface=\"Vrinda\"/>"
                "<a:font script=\"Gujr\" typeface=\"Shruti\"/>"
                "<a:font script=\"Khmr\" typeface=\"MoolBoran\"/>"
                "<a:font script=\"Knda\" typeface=\"Tunga\"/>"
                "<a:font script=\"Guru\" typeface=\"Raavi\"/>"
                "<a:font script=\"Cans\" typeface=\"Euphemia\"/>"
                "<a:font script=\"Cher\" typeface=\"Plantagenet Cherokee\"/>"
                "<a:font script=\"Yiii\" typeface=\"Microsoft Yi Baiti\"/>"
                "<a:font script=\"Tibt\" typeface=\"Microsoft Himalaya\"/>"
                "<a:font script=\"Thaa\" typeface=\"MV Boli\"/>"
                "<a:font script=\"Deva\" typeface=\"Mangal\"/>"
                "<a:font script=\"Telu\" typeface=\"Gautami\"/>"
                "<a:font script=\"Taml\" typeface=\"Latha\"/>"
                "<a:font script=\"Syrc\" typeface=\"Estrangelo Edessa\"/>"
                "<a:font script=\"Orya\" typeface=\"Kalinga\"/>"
                "<a:font script=\"Mlym\" typeface=\"Kartika\"/>"
                "<a:font script=\"Laoo\" typeface=\"DokChampa\"/>"
                "<a:font script=\"Sinh\" typeface=\"Iskoola Pota\"/>"
                "<a:font script=\"Mong\" typeface=\"Mongolian Baiti\"/>"
                "<a:font script=\"Viet\" typeface=\"Times New Roman\"/>"
                "<a:font script=\"Uigh\" typeface=\"Microsoft Uighur\"/>"
                "<a:font script=\"Geor\" typeface=\"Sylfaen\"/>"
                "<a:font script=\"Armn\" typeface=\"Arial\"/>"
                "<a:font script=\"Bugi\" typeface=\"Leelawadee UI\"/>"
                "<a:font script=\"Bopo\" typeface=\"Microsoft JhengHei\"/>"
                "<a:font script=\"Java\" typeface=\"Javanese Text\"/>"
                "<a:font script=\"Lisu\" typeface=\"Segoe UI\"/>"
                "<a:font script=\"Mymr\" typeface=\"Myanmar Text\"/>"
                "<a:font script=\"Nkoo\" typeface=\"Ebrima\"/>"
                "<a:font script=\"Olck\" typeface=\"Nirmala UI\"/>"
                "<a:font script=\"Osma\" typeface=\"Ebrima\"/>"
                "<a:font script=\"Phag\" typeface=\"Phagspa\"/>"
                "<a:font script=\"Syrn\" typeface=\"Estrangelo Edessa\"/>"
                "<a:font script=\"Syrj\" typeface=\"Estrangelo Edessa\"/>"
                "<a:font script=\"Syre\" typeface=\"Estrangelo Edessa\"/>"
                "<a:font script=\"Sora\" typeface=\"Nirmala UI\"/>"
                "<a:font script=\"Tale\" typeface=\"Microsoft Tai Le\"/>"
                "<a:font script=\"Talu\" typeface=\"Microsoft New Tai Lue\"/>"
                "<a:font script=\"Tfng\" typeface=\"Ebrima\"/></a:majorFont>"
                "<a:minorFont>"
                "<a:latin typeface=\"Calibri\" panose=\"020F0502020204030204\"/>"
                "<a:ea typeface=\"\"/>"
                "<a:cs typeface=\"\"/>"
                "<a:font script=\"Jpan\" typeface=\"\xe6\xb8\xb8\xe6\x98\x8e\xe6\x9c\x9d\"/>"
                "<a:font script=\"Hang\" typeface=\"\xeb\xa7\x91\xec\x9d\x80 \xea\xb3\xa0\xeb\x94\x95\"/>"
                "<a:font script=\"Hans\" typeface=\"\xe7\xad\x89\xe7\xba\xbf\"/>"
                "<a:font script=\"Hant\" typeface=\"\xe6\x96\xb0\xe7\xb4\xb0\xe6\x98\x8e\xe9\xab\x94\"/>"
                "<a:font script=\"Arab\" typeface=\"Arial\"/>"
                "<a:font script=\"Hebr\" typeface=\"Arial\"/>"
                "<a:font script=\"Thai\" typeface=\"Cordia New\"/>"
                "<a:font script=\"Ethi\" typeface=\"Nyala\"/>"
                "<a:font script=\"Beng\" typeface=\"Vrinda\"/>"
                "<a:font script=\"Gujr\" typeface=\"Shruti\"/>"
                "<a:font script=\"Khmr\" typeface=\"DaunPenh\"/>"
                "<a:font script=\"Knda\" typeface=\"Tunga\"/>"
                "<a:font script=\"Guru\" typeface=\"Raavi\"/>"
                "<a:font script=\"Cans\" typeface=\"Euphemia\"/>"
                "<a:font script=\"Cher\" typeface=\"Plantagenet Cherokee\"/>"
                "<a:font script=\"Yiii\" typeface=\"Microsoft Yi Baiti\"/>"
                "<a:font script=\"Tibt\" typeface=\"Microsoft Himalaya\"/>"
                "<a:font script=\"Thaa\" typeface=\"MV Boli\"/>"
                "<a:font script=\"Deva\" typeface=\"Mangal\"/>"
                "<a:font script=\"Telu\" typeface=\"Gautami\"/>"
                "<a:font script=\"Taml\" typeface=\"Latha\"/>"
                "<a:font script=\"Syrc\" typeface=\"Estrangelo Edessa\"/>"
                "<a:font script=\"Orya\" typeface=\"Kalinga\"/>"
                "<a:font script=\"Mlym\" typeface=\"Kartika\"/>"
                "<a:font script=\"Laoo\" typeface=\"DokChampa\"/>"
                "<a:font script=\"Sinh\" typeface=\"Iskoola Pota\"/>"
                "<a:font script=\"Mong\" typeface=\"Mongolian Baiti\"/>"
                "<a:font script=\"Viet\" typeface=\"Arial\"/>"
                "<a:font script=\"Uigh\" typeface=\"Microsoft Uighur\"/>"
                "<a:font script=\"Geor\" typeface=\"Sylfaen\"/>"
                "<a:font script=\"Armn\" typeface=\"Arial\"/>"
                "<a:font script=\"Bugi\" typeface=\"Leelawadee UI\"/>"
                "<a:font script=\"Bopo\" typeface=\"Microsoft JhengHei\"/>"
                "<a:font script=\"Java\" typeface=\"Javanese Text\"/>"
                "<a:font script=\"Lisu\" typeface=\"Segoe UI\"/>"
                "<a:font script=\"Mymr\" typeface=\"Myanmar Text\"/>"
                "<a:font script=\"Nkoo\" typeface=\"Ebrima\"/>"
                "<a:font script=\"Olck\" typeface=\"Nirmala UI\"/>"
                "<a:font script=\"Osma\" typeface=\"Ebrima\"/>"
                "<a:font script=\"Phag\" typeface=\"Phagspa\"/>"
                "<a:font script=\"Syrn\" typeface=\"Estrangelo Edessa\"/>"
                "<a:font script=\"Syrj\" typeface=\"Estrangelo Edessa\"/>"
                "<a:font script=\"Syre\" typeface=\"Estrangelo Edessa\"/>"
                "<a:font script=\"Sora\" typeface=\"Nirmala UI\"/>"
                "<a:font script=\"Tale\" typeface=\"Microsoft Tai Le\"/>"
                "<a:font script=\"Talu\" typeface=\"Microsoft New Tai Lue\"/>"
                "<a:font script=\"Tfng\" typeface=\"Ebrima\"/></a:minorFont></a:fontScheme>"
                "<a:fmtScheme name=\"Office\">"
                "<a:fillStyleLst>"
                "<a:solidFill>"
                "<a:schemeClr val=\"phClr\"/></a:solidFill>"
                "<a:gradFill rotWithShape=\"1\">"
                "<a:gsLst>"
                "<a:gs pos=\"0\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:lumMod val=\"110000\"/>"
                "<a:satMod val=\"105000\"/>"
                "<a:tint val=\"67000\"/></a:schemeClr></a:gs>"
                "<a:gs pos=\"50000\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:lumMod val=\"105000\"/>"
                "<a:satMod val=\"103000\"/>"
                "<a:tint val=\"73000\"/></a:schemeClr></a:gs>"
                "<a:gs pos=\"100000\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:lumMod val=\"105000\"/>"
                "<a:satMod val=\"109000\"/>"
                "<a:tint val=\"81000\"/></a:schemeClr></a:gs></a:gsLst>"
                "<a:lin ang=\"5400000\" scaled=\"0\"/></a:gradFill>"
                "<a:gradFill rotWithShape=\"1\">"
                "<a:gsLst>"
                "<a:gs pos=\"0\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:satMod val=\"103000\"/>"
                "<a:lumMod val=\"102000\"/>"
                "<a:tint val=\"94000\"/></a:schemeClr></a:gs>"
                "<a:gs pos=\"50000\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:satMod val=\"110000\"/>"
                "<a:lumMod val=\"100000\"/>"
                "<a:shade val=\"100000\"/></a:schemeClr></a:gs>"
                "<a:gs pos=\"100000\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:lumMod val=\"99000\"/>"
                "<a:satMod val=\"120000\"/>"
                "<a:shade val=\"78000\"/></a:schemeClr></a:gs></a:gsLst>"
                "<a:lin ang=\"5400000\" scaled=\"0\"/></a:gradFill></a:fillStyleLst>"
                "<a:lnStyleLst>"
                "<a:ln w=\"6350\" cap=\"flat\" cmpd=\"sng\" algn=\"ctr\">"
                "<a:solidFill>"
                "<a:schemeClr val=\"phClr\"/></a:solidFill>"
                "<a:prstDash val=\"solid\"/>"
                "<a:miter lim=\"800000\"/></a:ln>"
                "<a:ln w=\"12700\" cap=\"flat\" cmpd=\"sng\" algn=\"ctr\">"
                "<a:solidFill>"
                "<a:schemeClr val=\"phClr\"/></a:solidFill>"
                "<a:prstDash val=\"solid\"/>"
                "<a:miter lim=\"800000\"/></a:ln>"
                "<a:ln w=\"19050\" cap=\"flat\" cmpd=\"sng\" algn=\"ctr\">"
                "<a:solidFill>"
                "<a:schemeClr val=\"phClr\"/></a:solidFill>"
                "<a:prstDash val=\"solid\"/>"
                "<a:miter lim=\"800000\"/></a:ln></a:lnStyleLst>"
                "<a:effectStyleLst>"
                "<a:effectStyle>"
                "<a:effectLst/></a:effectStyle>"
                "<a:effectStyle>"
                "<a:effectLst/></a:effectStyle>"
                "<a:effectStyle>"
                "<a:effectLst>"
                "<a:outerShdw blurRad=\"57150\" dist=\"19050\" dir=\"5400000\" algn=\"ctr\" rotWithShape=\"0\">"
                "<a:srgbClr val=\"000000\">"
                "<a:alpha val=\"63000\"/></a:srgbClr></a:outerShdw></a:effectLst></a:effectStyle></a:effectStyleLst>"
                "<a:bgFillStyleLst>"
                "<a:solidFill>"
                "<a:schemeClr val=\"phClr\"/></a:solidFill>"
                "<a:solidFill>"
                "<a:schemeClr val=\"phClr\">"
                "<a:tint val=\"95000\"/>"
                "<a:satMod val=\"170000\"/></a:schemeClr></a:solidFill>"
                "<a:gradFill rotWithShape=\"1\">"
                "<a:gsLst>"
                "<a:gs pos=\"0\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:tint val=\"93000\"/>"
                "<a:satMod val=\"150000\"/>"
                "<a:shade val=\"98000\"/>"
                "<a:lumMod val=\"102000\"/></a:schemeClr></a:gs>"
                "<a:gs pos=\"50000\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:tint val=\"98000\"/>"
                "<a:satMod val=\"130000\"/>"
                "<a:shade val=\"90000\"/>"
                "<a:lumMod val=\"103000\"/></a:schemeClr></a:gs>"
                "<a:gs pos=\"100000\">"
                "<a:schemeClr val=\"phClr\">"
                "<a:shade val=\"63000\"/>"
                "<a:satMod val=\"120000\"/></a:schemeClr></a:gs></a:gsLst>"
                "<a:lin ang=\"5400000\" scaled=\"0\"/></a:gradFill></a:bgFillStyleLst></a:fmtScheme></a:themeElements>"
                "<a:objectDefaults/>"
                "<a:extraClrSchemeLst/>"
                "<a:extLst>"
                "<a:ext uri=\"{05A4C25C-085E-4340-85A3-A5531E510DB2}\">"
                "<thm15:themeFamily xmlns:thm15=\"http://schemas.microsoft.com/office/thememl/2012/main\" name=\"Office Theme\" id=\"{62F939B6-93AF-4DB8-9C6B-D6C7DFDC589F}\" vid=\"{4A3C46E8-61CC-4603-A589-7422A47A8E4A}\"/></a:ext></a:extLst></a:theme>"
    },

};

int docx_template_items_num = 11;
